from notbadai_ide import File

FILE_TYPE = {
    '.py': 'python',
    '.ts': 'typescript',
    '.js': 'javascript',
}

LINE_COMMENT = {
    'python': '#',
    'typescript': '//',
}


def markdown_code_block(content: str, *, type_: str = '') -> str:
    return f"```{type_}\n{content}\n```"


def markdown_section(title: str, content: str) -> str:
    return f"## {title}\n\n{content.strip()}"  # test


def add_line_comment(file: 'File', line: str, comment: str) -> str:
    ft = FILE_TYPE.get(file.suffix(), '')
    cm = LINE_COMMENT.get(ft, '//')

    return line.rstrip() + f'  {cm} {comment.strip()}'



import re

from notbadai_ide import api


def _strip_ansi(text):
    # ansi_escape = re.compile(r'(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]')
    ansi_escape = re.compile(r'\x1b\[[0-?]*[ -/]*[@-~]')
    return ansi_escape.sub('', text)


def _clean_carriage_return(text):
    """
    For terminal history

    Cleans terminal output containing carriage returns by removing overwritten lines.

    Args:
        text (str): Raw terminal output containing \r characters

    Returns:
        str: Cleaned text with overwritten lines removed
    """
    lines = text.split('\n')
    cleaned_lines = []

    for line in lines:
        # Split by carriage return and keep only the last segment
        segments = line.split('\r')
        line = []
        for s in segments:
            s = list(s)
            if len(line) > len(s):
                line[:len(s)] = s
            else:
                line = s

        line = ''.join(line)
        if segments:
            cleaned_lines.append(line.rstrip())

    return '\n'.join(cleaned_lines)


def _clean_empty_lines(text):
    lines = text.split('\n')

    while lines and lines[0].strip() == '':
        lines.pop(0)
    while lines and lines[-1].strip() == '':
        lines.pop()

    return '\n'.join(lines)


def get_terminal_snapshot():
    terminal_snapshot = api.get_current_terminal().get_snapshot()
    api.log(terminal_snapshot)
    text = _strip_ansi('\n'.join(terminal_snapshot))
    return _clean_empty_lines(text)



import re
from typing import Tuple, List

from labml import monit


def compare_line(l1, l2):
    l1, s1 = l1
    l2, s2 = l2

    if len(l1) == 0 and len(l2) == 0:
        return 0.0

    s = len(s1 & s2) / max(len(s1), len(s2))
    if s < 0.2:
        return -1.
    elif s < 0.5:
        return s - 1.
    else:
        pass
        # return s

    d = [[0 for _ in range(len(l2) + 1)] for _ in range(len(l1) + 1)]
    for i in range(len(l1) - 1, -1, -1):
        for j in range(len(l2) - 1, -1, -1):
            d[i][j] = max(
                d[i + 1][j],
                d[i][j + 1],
                d[i + 1][j + 1] + (len(l1[i]) if l1[i] == l2[j] else 0),
            )

    s = d[0][0] / max(len(''.join(l1)), len(''.join(l2)))
    if s > 0.5:
        return s
    else:
        return s - 1.


def split_string(s):
    pattern = r'[a-zA-Z0-9]+|[^\s]'
    return re.findall(pattern, s)


def compress_line(line):
    parts = split_string(line)
    return parts, set(parts)


def get_matches(v1: str, v2: str) -> Tuple[List[List[int]], List[str]]:
    v1, v2 = v1.splitlines(keepends=True), v2.splitlines(keepends=True)

    sv1 = [compress_line(line) for line in v1]
    sv2 = [compress_line(line) for line in v2]

    diff = [[compare_line(l1, l2) for l2 in sv2] for l1 in monit.iterate(sv1)]

    dp = [[0 for _ in range(len(v2) + 1)] for _ in range(len(v1) + 1)]

    for i in range(len(v1) - 1, -1, -1):
        for j in range(len(v2) - 1, -1, -1):
            dp[i][j] = max(
                dp[i + 1][j],
                dp[i][j + 1],
                dp[i + 1][j + 1] + diff[i][j],
            )

    matches = []
    i, j = 0, 0
    while i < len(v1) and j < len(v2):
        if dp[i][j] == dp[i + 1][j + 1] + diff[i][j]:
            matches.append((i, j))
            i += 1
            j += 1
        elif dp[i][j] == dp[i + 1][j]:
            i += 1
        elif dp[i][j] == dp[i][j + 1]:
            j += 1
        else:
            raise RuntimeError()

    matches.append([len(v1), len(v2)])

    return matches, v2



LLM_PROVIDERS = [
    {
        'name': 'deepinfra',
        'base_url': 'https://api.deepinfra.com/v1/openai',
    },
    {
        'name': 'openrouter',
        'base_url': 'https://openrouter.ai/api/v1',
    }
]



FILE_TYPE_MAP = {
    '.py': 'python',
    '.js': 'javascript',
    '.ts': 'typescript',
    '.jsx': 'javascript',
    '.tsx': 'typescript',
    '.java': 'java',
    '.c': 'c',
    '.cpp': 'cpp',
    '.cc': 'cpp',
    '.cxx': 'cpp',
    '.h': 'c',
    '.hpp': 'cpp',
    '.cs': 'csharp',
    '.php': 'php',
    '.rb': 'ruby',
    '.go': 'go',
    '.rs': 'rust',
    '.swift': 'swift',
    '.kt': 'kotlin',
    '.scala': 'scala',
    '.sh': 'bash',
    '.bash': 'bash',
    '.zsh': 'zsh',
    '.fish': 'fish',
    '.html': 'html',
    '.htm': 'html',
    '.css': 'css',
    '.scss': 'scss',
    '.sass': 'sass',
    '.less': 'less',
    '.json': 'json',
    '.xml': 'xml',
    '.yaml': 'yaml',
    '.yml': 'yaml',
    '.toml': 'toml',
    '.ini': 'ini',
    '.cfg': 'ini',
    '.conf': 'conf',
    '.md': 'markdown',
    '.markdown': 'markdown',
    '.txt': 'text',
    '.sql': 'sql',
    '.r': 'r',
    '.R': 'r',
    '.m': 'matlab',
    '.pl': 'perl',
    '.lua': 'lua',
    '.vim': 'vim',
    '.dockerfile': 'dockerfile',
    '.makefile': 'makefile',
}


def get_file_type(file_path: str) -> str:
    """Determine file type based on file extension."""
    if not file_path:
        return 'unknown'

    # Handle special cases first
    filename = file_path.lower().split('/')[-1]
    if filename in ('dockerfile', 'makefile'):
        return filename

    # check for file extension
    if '.' in filename:
        ext = '.' + filename.split('.')[-1]
        return FILE_TYPE_MAP.get(ext, 'unknown')

    return 'unknown'



# Default model mappings
MODELS = {
    'v3': {
        'openrouter': 'deepseek/deepseek-v3.2-exp',
        'deepinfra': 'deepseek-ai/DeepSeek-V3-0324-Turbo'
    },
    'devstral': {
        'openrouter': 'mistralai/devstral-medium',
        'deepinfra': 'mistralai/Devstral-Small-2507'
    },
    'default': {
        'deepinfra': 'Qwen/Qwen3-Coder-480B-A35B-Instruct-Turbo',
        'openrouter': 'anthropic/claude-sonnet-4.5'
    },
    'qwen': {
        'deepinfra': 'Qwen/Qwen3-Coder-480B-A35B-Instruct-Turbo',
        'openrouter': 'qwen/qwen3-coder'
    },
    'morph_fast': {
        'openrouter': 'morph/morph-v3-fast'
    },
    'morph_large': {
        'openrouter': 'morph/morph-v3-large'
    }
}



import time
from typing import Dict, List

from openai import OpenAI
from notbadai_ide import api, START_METADATA, END_METADATA, START_THINK, END_THINK

from .models import MODELS
from .settings import LLM_PROVIDERS


def call_llm(model_id: str,
             messages: List[Dict[str, str]],
             *,
             push_to_chat: bool = True,
             temperature: float = 1.0,
             top_p: float = 1.0,
             n_outputs: int = 1,
             max_tokens: int = None,
             ):
    """Streams responses from the LLM and sends them to the chat UI in real-time."""

    model_info = MODELS[model_id]

    api_keys = api.get_api_keys()
    if len(api_keys) == 0:
        raise ValueError('API provider required. Configure at least one in Extensions → Management.')

    default_key = None
    for k in api_keys:
        if k.default:
            default_key = k

    api_key, model_name = None, None
    if default_key.provider in model_info:
        model_name = model_info[default_key.provider]
        api_key = default_key
    else:
        for k in api_keys:
            if k.provider in model_info:
                model_name = model_info[k.provider]
                api_key = k
                break

    if api_key is None:
        raise ValueError(f"The API provider does not support {model_id} model")

    provider = None
    for p in LLM_PROVIDERS:
        if p['name'] == api_key.provider:
            provider = p
            break

    start_time = time.time()

    client = OpenAI(api_key=api_key.key, base_url=provider['base_url'])

    stream = client.chat.completions.create(
        model=model_name,
        messages=messages,
        stream=True,
        temperature=temperature,
        top_p=top_p,
        n=n_outputs,
        max_tokens=max_tokens,
    )

    thinking = False
    usage = None
    content = ''

    for chunk in stream:
        delta = chunk.choices[0].delta
        if push_to_chat:
            if getattr(delta, 'reasoning', None):
                if not thinking:
                    api.chat(START_THINK)
                    thinking = True
                api.chat(content=delta.reasoning)

        if delta.content:
            if push_to_chat:
                if thinking:
                    api.chat(END_THINK)
                    thinking = False
                api.chat(content=delta.content)
            content += delta.content

        if chunk.usage is not None:
            assert usage is None
            usage = chunk.usage

    elapsed = time.time() - start_time
    meta_data = f'Time: {elapsed:.2f}s'
    if usage is not None:
        api.log(str(usage))

        meta_data += f' Prompt tokens: {usage.prompt_tokens :,} Completion tokens {usage.completion_tokens :,}, Model: {model_name} @ {provider["name"]}'

    if push_to_chat:
        api.chat(f'{START_METADATA}{meta_data.strip()}{END_METADATA}')

        api.end_chat()

    return content



from git import Repo


class GitClient:
    def __init__(self, project_path: str):
        self.project_path = project_path
        self.repo = Repo(project_path)

    def is_repository(self) -> bool:
        """Check if the current directory is a git repository."""
        return not self.repo.bare

    def get_commit_diff(self, staged_only: bool = False) -> str:
        if staged_only:
            # Get diff of staged files only
            return self.repo.git.diff('--cached')
        else:
            # Get diff of all modified files (staged and unstaged)
            return self.repo.git.diff('HEAD')

    def commit_push(self, commit_message: str) -> None:
        """Commit all changes and push to the current branch."""
        # Stage all changes
        self.repo.git.add('-A')

        # Check if there are changes to commit
        if self.repo.is_dirty() or self.repo.untracked_files:
            # Commit changes
            commit = self.repo.index.commit(commit_message)

            # Get changed files - commit.stats.files.keys() already returns file paths as strings
            changed_files = list(commit.stats.files.keys())

            # Push to current branch
            origin = self.repo.remote('origin')
            current_branch = self.repo.active_branch.name

            # Try to push with upstream
            origin.push(f"{current_branch}:{current_branch}", set_upstream=True)


import inspect
import json
import re
import typing
from pathlib import Path
from string import Template
from typing import Optional

from notbadai_ide import api


def parse_prompt() -> typing.Tuple[str, str, str]:
    prompt = api.get_prompt().strip()

    if prompt.startswith('@'):
        model = prompt.split()[0][1:]  # get text after @
        prompt = prompt[len(model) + 1:].strip()  # remove @model from prompt
    else:
        model = 'default'

    if prompt.startswith('\\'):
        command = prompt.split()[0][1:].strip()  # get text after \
        prompt = prompt[len(command) + 1:].strip()
    else:
        command = 'context'

    return command, model, prompt


def add_line_numbers(code: str) -> str:
    res = [str(idx + 1) + ": " + line.rstrip() for idx, line in enumerate(code.splitlines())]
    return '\n'.join(res)


def get_prompt_template(template_path: str, **kwargs) -> str:
    frame = inspect.stack()[1]
    # Get the file path of the caller
    caller_file = frame.filename

    path = Path(caller_file).parent / f'{template_path}.md'
    with open(str(path)) as f:
        template = Template(f.read())

    return template.substitute(kwargs)


def extract_code_block(text: str, language: Optional[str] = None, ignore_no_ticks=False) -> Optional[str]:
    if language:
        pattern = rf"```{re.escape(language)}\s*\n(.*?)```"
    else:
        pattern = r"```(?:\w+)?\s*\n(.*?)```"

    m = re.search(pattern, text, re.DOTALL)

    if m:
        return m.group(1)

    if ignore_no_ticks:
        return text
    else:
        return None


def parse_json(response: str):
    try:
        return json.loads(response)
    except json.decoder.JSONDecodeError:
        block = extract_code_block(response)
        block = block.strip()
        if not block:
            raise ValueError("No JSON block detected in model response")
        try:
            return json.loads(block)
        except json.decoder.JSONDecodeError as e:
            api.log(block)
            raise e



from typing import List
from notbadai_ide import api, File, START_METADATA, END_METADATA

from .terminal import get_terminal_snapshot
from .formatting import markdown_section, markdown_code_block, add_line_comment


def build_context(files: List['File'] = None) -> str:
    """Builds the context string from the current file and selection."""
    context = []

    current_file = api.get_current_file()
    context_files = api.get_context_files()
    selection = api.get_selection()
    terminal = get_terminal_snapshot()
    cursor = api.get_cursor()

    other_files = []
    if files is None:
        file_list = api.get_repo_files()
        for file in file_list:
            if file.is_open:
                other_files.append(file)
    else:
        file_list = files
        other_files = file_list

    if context_files:
        api.chat(f'{START_METADATA}Context paths: {", ".join(list(context_files.keys()))}{END_METADATA}')

    if file_list:
        repo_files = [f'{f.path}`' for f in file_list]
        context.append(markdown_section("List of Files", "\n".join(repo_files)))

    if other_files:
        api.chat(f'{START_METADATA}Opened files: {", ".join(f.path for f in other_files)}{END_METADATA}')

    # combine other_files with files from context_files, removing duplicates
    all_files = (other_files or []) + [f for files_list in context_files.values() for f in files_list]
    relevant_files = list({f.path: f for f in all_files}.values())

    if relevant_files:
        api.chat(f'{START_METADATA}Relevant files: {", ".join(f.path for f in relevant_files)}{END_METADATA}')
        relevant_files = [f'Path: `{f.path}`\n\n' + markdown_code_block(f.get_content()) for f in relevant_files]
        context.append(markdown_section("Relevant files", "\n\n".join(relevant_files)))

    if current_file:
        api.chat(f'{START_METADATA}Current file: {current_file.path}{END_METADATA}')
        context.append(
            markdown_section("Current File",
                             f"Path: `{current_file.path}`\n\n" +
                             markdown_code_block(current_file.get_content()))
        )

    if terminal:
        if len(terminal) > 40_000:
            pre_text = f'Terminal output is {len(terminal)} chars long, and here is the last 40k chars of it.\n\n'
        else:
            pre_text = f'Terminal output is {len(terminal)} chars long.'

        context.append(
            markdown_section("Terminal output",
                             f"{pre_text}\n\n" + markdown_code_block(terminal[-40000:]))
        )

    if selection and selection.strip():
        context.append(
            markdown_section("Selection",
                             "This is the code snippet that I'm referring to\n\n" +
                             markdown_code_block(selection))
        )

    if current_file and cursor:
        cursor_row, cursor_column = cursor.row - 1, cursor.column - 1
        block = current_file.get_content().split('\n')
        assert len(block) > cursor_row, f'Cursor row {cursor_row} block of length {len(block)}'
        prefix = block[cursor_row - 3: cursor_row]
        line = block[cursor_row]
        line = add_line_comment(current_file, line, f'Cursor is here: `{line[:cursor_column].strip()}`')
        suffix = block[cursor_row + 1:cursor_row + 4]

        block = prefix + [line] + suffix

        context.append(markdown_section("Cursor position",
                                        markdown_code_block('\n'.join(block))))

    return "\n\n".join(context)
